# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Key Binding Action Handler Shim"""

from functools import update_wrapper

__all__ = (
    'KeyActionMap',
)


class KeyActionMap(object):
    """"""
    def __init__(self, carousel):
        self.carousel = carousel

        self.zone_manager = carousel.zone_manager

        self.zone_content = carousel.zone_manager.zone_content
        self.zone_details = carousel.zone_manager.zone_details
        self.zone_lowdown = carousel.zone_manager.zone_lowdown

        self.update_handler = carousel.update_handler

    # ***

    class Decorators(object):
        @classmethod
        def debug_log_trace_enter_leave(cls, func):
            def trace_enter_leave_wrapper(obj, event, *args, **kwargs):
                # 2019-01-17: [lb]: I added this wrapper to help delimit debug
                # trace messages (to determine where each command's messages
                # begin and end). But it might later be useful for other tasks,
                # such as profiling. So leaving here, but with a note that says,
                # yeah, this code has little utility to the end consumer, other
                # than to make the developer more comfortable in the code jungle.
                debug = obj.carousel.controller.client_logger.debug
                debug('🚿 🐎 ENTER 👋 🍩 “{}”'.format(func.__name__))
                func(obj, event, *args, **kwargs)
                # Include a visual delimiter to make it easy to scan log trace
                # and see groups of messages belonging to each command.
                debug('🍖 🛀 LEAVE 🐵 🍌 “{}”'.format(func.__name__))

            return update_wrapper(trace_enter_leave_wrapper, func)

        # ***

        @classmethod
        def refresh_now(cls, func):
            def wrapper(obj, event, *args, **kwargs):
                # So that we don't have to call controller.store.now_tz_aware()
                # in our handlers, and so our handlers do not have to worry that
                # "now" has different values during the event handling, reset now
                # now.
                obj.carousel.controller.now_refresh()
                func(obj, event, *args, **kwargs)

            return update_wrapper(wrapper, func)

        # ***

        @classmethod
        def intercept_modifier(cls, reset=False):
            """Passes key press to command modifier accumulator if necessary.

            Also resets the command modifier if indicatd by the command being
            decorated."""
            def _intercept_modifier(func):
                def wrapper(obj, event, *args, **kwargs):
                    command_modifier = obj.carousel.update_handler.command_modifier
                    re_date_seps = obj.carousel.update_handler.re_date_seps
                    # Check if '.' and pass to command modifier if started.
                    # - This enables a fractional command modifier to coexist
                    #   with a user command mapped to '.'. I.e., if no modifier
                    #   and user presses '.', execute the mapped command. But
                    #   if modifier started and '.', treat as decimal point.
                    if (
                        command_modifier
                        and (
                            (
                                event.data == '.'
                                # Can only be one '.' in modifier, so check not seen.
                                # (lb): This is somewhat awkward behavior, e.g., if
                                # user tries typing a date prefix (e.g., for the 'G'
                                # command) using periods, say, `2020.01.` on the
                                # second dot, it'll execute the '.' command. Oh well.
                                and '.' not in command_modifier
                            )
                            or (
                                # Note that date separators are allowed more than once,
                                # meaning any command mapped to a key also used as a
                                # date separator will be shadowed by this code and
                                # therefore cannot be called with a command modifier.
                                re_date_seps.match(event.data)
                            )
                        )
                    ):
                        obj.carousel.update_handler.command_modifier_any_key(event)
                    else:
                        if reset:
                            obj.carousel.update_handler.reset_time_multipliers()
                        func(obj, event, *args, **kwargs)

                return update_wrapper(wrapper, func)

            return _intercept_modifier

        # ***

    # #### Key bindings wired by KeyBonder.widget_focus().

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def focus_next(self, event):
        self.zone_manager.focus_next(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def focus_previous(self, event):
        self.zone_manager.focus_previous(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def edit_time_start(self, event):
        self.zone_manager.toggle_focus_time_start(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def edit_time_end(self, event):
        self.zone_manager.toggle_focus_time_end(event)

    # #### Key bindings wired by KeyBonder.save_and_quit().

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def save_edited_and_live(self, event):
        self.carousel.save_edited_and_live(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def save_edited_and_exit(self, event):
        self.carousel.save_edited_and_exit(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def exit_command(self, event):
        self.carousel.exit_command(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def exit_quietly(self, event):
        was_helping = self.zone_content.on_reset_hide_help()
        if was_helping:
            return
        self.carousel.exit_quietly(event)

    # #### Key bindings wired by KeyBonder.edit_time().

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def edit_time_enter(self, event):
        self.zone_details.edit_time_enter(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def toggle_focus_description(self, event):
        self.zone_details.toggle_focus_description(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def edit_time_any_key(self, event):
        self.zone_details.edit_time_any_key(event)

    # #### Key bindings wired by KeyBonder.undo_redo().

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def undo_command_content(self, event):
        self.update_handler.undo_command(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def redo_command_content(self, event):
        self.update_handler.redo_command(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def undo_command_edit_time(self, event):
        self.zone_details.undo_command(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def redo_command_edit_time(self, event):
        self.zone_details.redo_command(event)

    # #### Key bindings wired by KeyBonder.shortcuts().

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def copy_complete_and_paste_active(self, event):
        self.update_handler.copy_complete_and_paste_active(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def copy_complete_and_paste_new(self, event):
        self.update_handler.copy_complete_and_paste_new(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def complete_and_prompt_new(self, event):
        self.update_handler.complete_and_prompt_new(event)

    # #### Key bindings wired by KeyBonder.custom_factoids().

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def custom_factoid_paste(self, event):
        self.update_handler.custom_factoid_paste(event)

    # #### Key bindings wired by KeyBonder.normal().

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def rotate_help(self, event):
        self.zone_content.rotate_help(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def dev_breakpoint(self, event):
        self.carousel.dev_breakpoint(event)

    # *** Next/Prev: Fact

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier()
    def jump_fact_dec(self, event):
        self.zone_manager.jump_fact_dec(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier()
    def jump_fact_inc(self, event):
        try:
            self.zone_manager.jump_fact_inc(event)
        except Exception:
            # 2019-12-03 01:18: Raised on add-tag then save. Don't remember jumping.
            self.carousel.controller.affirm(False)
            pass

    # *** Next/Prev: Day

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier()
    def jump_day_dec(self, event):
        self.zone_manager.jump_day_dec(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier()
    def jump_day_inc(self, event):
        self.zone_manager.jump_day_inc(event)

    # *** Next/Prev: Rift

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier()
    def jump_rift_dec(self, event):
        self.zone_manager.jump_rift_dec(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier()
    def jump_rift_inc(self, event):
        self.zone_manager.jump_rift_inc(event)

    # *** First/Final: Fact

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def jump_fact_first(self, event):
        self.zone_manager.jump_fact_first(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def jump_fact_final(self, event):
        self.zone_manager.jump_fact_final(event)

    # *** Up/Down: Content Cursor Motion

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier()
    def cursor_up_one(self, event):
        self.zone_content.cursor_up_one(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier()
    def cursor_down_one(self, event):
        self.zone_content.cursor_down_one(event)

    # *** Up/Down/Top/Bottom: Content Scrolling

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def scroll_up(self, event):
        self.zone_content.scroll_up(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def scroll_down(self, event):
        self.zone_content.scroll_down(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def scroll_top(self, event):
        self.zone_content.scroll_top(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def scroll_bottom(self, event):
        self.zone_content.scroll_bottom(event)

    # #### Key bindings wired by KeyBonder.edit_fact().

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def edit_fact(self, event):
        self.update_handler.edit_fact(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def edit_actegory(self, event):
        self.update_handler.edit_actegory(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def edit_description(self, event):
        self.update_handler.edit_description(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def edit_tags(self, event):
        self.update_handler.edit_tags(event)

    # #### Key bindings wired by KeyBonder.nudge_time().

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier()
    def edit_time_decrement_start(self, event):
        self.update_handler.edit_time_decrement_start(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier()
    def edit_time_increment_start(self, event):
        self.update_handler.edit_time_increment_start(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier()
    def edit_time_decrement_end(self, event):
        self.update_handler.edit_time_decrement_end(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier()
    def edit_time_increment_end(self, event):
        self.update_handler.edit_time_increment_end(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier()
    def edit_time_decrement_both(self, event):
        self.update_handler.edit_time_decrement_both(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier()
    def edit_time_increment_both(self, event):
        self.update_handler.edit_time_increment_both(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier()
    def edit_time_decrement_start_5min(self, event):
        self.update_handler.edit_time_decrement_start_5min(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier()
    def edit_time_increment_start_5min(self, event):
        self.update_handler.edit_time_increment_start_5min(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier()
    def edit_time_decrement_end_5min(self, event):
        self.update_handler.edit_time_decrement_end_5min(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier()
    def edit_time_increment_end_5min(self, event):
        self.update_handler.edit_time_increment_end_5min(event)

    # #### Key bindings wired by KeyBonder.command_modifier().

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    def allow_time_gap(self, event):
        self.update_handler.allow_time_gap(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    # Not necessary: @Decorators.intercept_modifier()
    # (because gaits to the same handler anyway).
    def command_modifier_any_key(self, event):
        self.update_handler.command_modifier_any_key(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    # NOPE: @Decorators.intercept_modifier()
    def backspace_command_modifier(self, event):
        self.update_handler.backspace_command_modifier(event)

    # #### Key bindings wired by KeyBonder.create_delete_fact().

    # FIXME/2020-04-11: Remove these, or implement!

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def fact_split(self, event):
        self.update_handler.fact_split(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def fact_erase(self, event):
        self.update_handler.fact_erase(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def fact_merge_prev(self, event):
        self.update_handler.fact_merge_prev(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def fact_merge_next(self, event):
        self.update_handler.fact_merge_next(event)

    # #### Key bindings wired by KeyBonder.clipboard().

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def fact_copy_fact(self, event):
        self.update_handler.fact_copy_fact(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def fact_cut(self, event):
        self.update_handler.fact_cut(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def fact_paste(self, event):
        self.update_handler.fact_paste(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def fact_copy_activity(self, event):
        self.update_handler.fact_copy_activity(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def fact_copy_tags(self, event):
        self.update_handler.fact_copy_tags(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def fact_copy_description(self, event):
        self.update_handler.fact_copy_description(event)

    # #### Key bindings wired by KeyBonder.begin_commando().

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def begin_commando(self, event):
        self.update_handler.begin_commando(event)

    # #### Key bindings wired by KeyBonder.going_commando().

    # The following handlers occur after keybindings are rewired,
    # so there's no need to call the intercept decorator:
    #
    #   NOPE: @Decorators.intercept_modifier

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    def cancel_commando(self, event):
        self.update_handler.cancel_commando(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    def parts_commando(self, event):
        self.update_handler.parts_commando(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    def backspace_commando(self, event):
        self.update_handler.backspace_commando(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    def final_commando(self, event):
        self.update_handler.final_commando(event)

    # #### Key bindings wired by KeyBonder.begin_delta_time().

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def begin_delta_time_start(self, event):
        self.update_handler.begin_delta_time_start(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    @Decorators.intercept_modifier(reset=True)
    def begin_delta_time_end(self, event):
        self.update_handler.begin_delta_time_end(event)

    # #### Key bindings wired by KeyBonder.going_delta_time().

    # The following handlers occur after keybindings are rewired,
    # so there's no need to call the intercept decorator:
    #
    #   NOPE: @Decorators.intercept_modifier

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    def cancel_delta_time(self, event):
        self.update_handler.cancel_delta_time(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    def parts_delta_time(self, event):
        self.update_handler.parts_delta_time(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    def backspace_delta_time(self, event):
        self.update_handler.backspace_delta_time(event)

    # Elsewhere: allow_time_gap (Wired by KeyBonder.command_modifier, too).

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    def final_delta_time_apply(self, event):
        self.update_handler.final_delta_time_apply(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    def final_delta_time_minutes(self, event):
        self.update_handler.final_delta_time_minutes(event)

    @Decorators.debug_log_trace_enter_leave
    @Decorators.refresh_now
    def final_delta_time_hours(self, event):
        self.update_handler.final_delta_time_hours(event)




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""dob_viewer.config sub.package provides Carousel UX user configuration settings."""

from gettext import gettext as _

from nark.config import ConfigRoot

__all__ = (
    'DobViewerConfigCustomPaste',
)


# ***

@ConfigRoot.section('custom-paste')
class DobViewerConfigCustomPaste(object):
    """"""

    def __init__(self, *args, **kwargs):
        pass

    # ***

    # (lb): This is the best I've got on a moment's thought:
    # Create placeholder config so that it'll be read from user's dob.conf.
    # (lb): This is not the cleanest solutions (we have to arbitrarily decide how
    # many placeholders to create), but it's otherwise transparent to the user.

    def add_custom_paste(postfix):
        """"""
        factoid_name = 'factoid_{}'.format(postfix)
        mapping_name = 'mapping_{}'.format(postfix)

        @property
        @ConfigRoot.setting(
            _("XXX"),
            name=factoid_name,
            hidden=True,
        )
        def factoid_prop(self):
            return ''

        @property
        @ConfigRoot.setting(
            _("XXX"),
            name=mapping_name,
            hidden=True,
        )
        def mapping_prop(self):
            return ''

    A_PERFECT_NUMBER = 28

    for postfix in range(1, A_PERFECT_NUMBER):
        add_custom_paste(postfix)




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Manages loading custom Lexer specified by user's config."""

from gettext import gettext as _

import inspect

import pygments.lexers
from prompt_toolkit.lexers import Lexer, PygmentsLexer

from dob_bright.styling import load_obj_from_internal
from dob_bright.termio import dob_in_user_warning

from ..ptkui import various_lexers

__all__ = (
    'load_content_lexer',
)


# If you want to test the lexers (ptkui/various_lexers.py),
# you can set your config, e.g.,
#   dob config set editor.lexer rainbow
#   dob config set editor.lexer truncater
#   dob config set editor.lexer wordwrapper
# or you can try it as a one-off with inline config 'tax:
#   dob -c editor.lexer=rainbow edit
#   dob -c editor.lexer=truncater edit
#   dob -c editor.lexer=wordwrapper edit

def load_content_lexer(controller):
    config = controller.config
    cfg_key_lexer = 'editor.lexer'

    def _load_content_lexer():
        named_lexer = resolve_named_lexer()
        lexer_class = load_obj_from_internal(
            controller,
            obj_name=named_lexer,
            internal=various_lexers,
            default_name=None,
            warn_tell_not_found=False,
            config_key=cfg_key_lexer,
        )
        return instantiate_or_try_pygments_lexer(named_lexer, lexer_class)

    def resolve_named_lexer():
        return config[cfg_key_lexer]

    def instantiate_or_try_pygments_lexer(named_lexer, lexer_class):
        if lexer_class is not None:
            return instantiate_class_maybe_from_method(lexer_class)
        return load_pygments_lexer(named_lexer)

    def instantiate_class_maybe_from_method(lexer_class):
        if inspect.isroutine(lexer_class):
            # (lb): This happens on say, `dob -c editor.lexer=wordwrapper edit`.
            # Also remember: ismethod for Class methods; isfunction for functions;
            # so use isroutine to check either.
            lexer_class = lexer_class()
        controller.affirm(inspect.isclass(lexer_class))
        content_lexer = lexer_class()
        controller.affirm(isinstance(content_lexer, Lexer))
        return content_lexer

    def load_pygments_lexer(named_lexer):
        # (lb): I'm a reSTie, personally, so we default to that.
        # (Though really the default is set in config/__init__.py.)
        lexer_name = named_lexer or 'RstLexer'
        try:
            return PygmentsLexer(getattr(pygments.lexers, lexer_name))
        except AttributeError:
            msg = _('Not a recognized Pygments lexer: “{0}”').format(lexer_name)
            dob_in_user_warning(msg)
            return None

    return _load_content_lexer()




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

""""""

from nark.items.fact import UntilTimeStops

__all__ = (
    'FactsManager_Jump',
)


class FactsManager_Jump(object):
    """"""

    # ***

    def fulfill_jump(self, jump_fact, reason=''):
        # Caller will have updated curr_group and curr_index, if necessary.
        # Note that we do not set FactsManager self.curr_fact directly,
        # but call a shim function so EditsManager can react, too.
        #   self.curr_fact = jump_fact
        self.on_jumped_fact(jump_fact)
        self._jump_time_reference = None
        self.debug_log_facts_mgr_state(reason)

    # ***

    def debug_log_facts_mgr_state(self, caller_name):
        self.debug(
            '{}: len(groups): {} / curr_index: {}\n'
            '- grps:\n{}\n- cgrp: {}\n- curr: {}'
            .format(
                caller_name,
                len(self.groups),
                self.curr_index,
                self.debug__str,
                self.curr_group,
                self.curr_fact.short,
            )
        )

    # ***

    def jump_to_fact_nearest(self, since_time=None, until_time=None):
        """"""
        def _jump_to_fact_nearest():
            assert (since_time is not None) ^ (until_time is not None)
            ref_time = since_time or until_time
            fact_group, group_fact, is_perfect = find_nearest_group_fact(ref_time)
            store_fact = is_perfect and None or find_nearest_store_fact(ref_time)
            nearest_fact = choose_nearest(fact_group, group_fact, store_fact)
            debug_log_chosen_fact(
                ref_time, nearest_fact, fact_group, group_fact, store_fact,
            )

            if nearest_fact is not None:
                if nearest_fact.pk not in self.by_pk.keys():
                    self.add_facts([nearest_fact])
                reason = 'jump-{}'.format('next' if since_time else 'prev')
                self.fulfill_jump(nearest_fact, reason=reason)
            update_time_reference(ref_time)

            return nearest_fact

    # ***

        def find_nearest_group_fact(ref_time):
            # Find the group index nearest ref_time. We use UntilTimeStops
            # so the index is 1 greater than group index for most matches
            # (all except the final ongoing (active) fact group).
            sorty_times = (ref_time, UntilTimeStops)
            inserts_at = self.groups.bisect_key_left(sorty_times)

            # Check for ongoing (active) fact group.
            if inserts_at == (len(self.groups) - 1):
                if sorty_times == self.groups[-1].sorty_times:
                    return self.groups[-1], self.groups[-1][0], True

            # If inserts_at is 0, ref_time is before any group's since_time,
            # or there's only one group and it's ongoing (active).
            if inserts_at == 0:
                first_group = self.groups[0]
                if ref_time == first_group.time_since:
                    return first_group, first_group[0], True
                elif since_time is not None:
                    # Momentum is forward, so grab first group's fact;
                    # and return False, so caller knows to look in store
                    # and decide between the two.
                    return first_group, first_group[0], False
                else:
                    # Momentum is backward in time, and there's nothing there.
                    return first_group, None, False

            try_group = self.groups[inserts_at - 1]
            self.controller.affirm(ref_time >= try_group.time_since)
            self.controller.affirm(sorty_times != try_group.sorty_times)

            # Check whether within group time window.
            if ref_time <= try_group.time_until:
                fact_index = try_group.bisect_key_left(sorty_times)
                best_fact = match_group_later_index(try_group, fact_index, ref_time)
                return try_group, best_fact, True

            # Between groups. Use momentum to determine which Fact to return.
            if since_time is not None:
                # Going forward.
                if inserts_at < len(self.groups):
                    # More groups to come!
                    return self.groups[inserts_at], self.groups[inserts_at][0], False
                else:
                    # On last group.
                    return self.groups[-1], None, False
            # Going backward. And we already processed inserts_at == 0.
            self.controller.affirm(until_time is not None)
            return try_group, try_group[-1], False

        # ***

        def match_group_later_index(fact_group, fact_index, ref_time):
            if fact_index >= len(fact_group):
                # Falls on or after start of last fact in group, but before
                # end of group time window, so use last fact of group.
                return fact_group[fact_index - 1]
            if fact_group[fact_index].start == ref_time:
                # Exact match. Use found fact.
                return fact_group[fact_index]
            # Falls before indicated fact.
            nearest_fact = fact_group[fact_index - 1]
            self.controller.affirm(nearest_fact.start <= ref_time)
            return nearest_fact

        # ***

        def find_nearest_store_fact(ref_time):
            if since_time is not None:
                return self.fetch_next_from_store(ref_time)
            else:
                return self.fetch_prev_from_store(ref_time)

        # ***

        def choose_nearest(fact_group, group_fact, store_fact):
            if group_fact is None:
                return choose_store_if_not_shadowed(fact_group, store_fact)
            if store_fact is None:
                return group_fact
            return prefer_store_then_group(fact_group, group_fact, store_fact)

        # ^^^

        def choose_store_if_not_shadowed(fact_group, store_fact):
            if store_fact is None:
                return None
            if since_time is not None:
                return store_if_not_shadowed_since(fact_group, None, store_fact)
            return store_if_not_shadowed_until(fact_group, None, store_fact)

        # ^^^

        def prefer_store_then_group(fact_group, group_fact, store_fact):
            if since_time is not None:
                return store_if_not_shadowed_until(fact_group, group_fact, store_fact)
            return store_if_not_shadowed_since(fact_group, group_fact, store_fact)

        # ^^^

        def store_if_not_shadowed_since(fact_group, group_fact, store_fact):
            if fact_group.time_until < store_fact.start:
                return store_fact
            return group_fact

        def store_if_not_shadowed_until(fact_group, group_fact, store_fact):
            store_fact_end = store_fact.end or UntilTimeStops
            if store_fact_end < fact_group.time_since:
                return store_fact
            return group_fact

        # ***

        def update_time_reference(ref_time):
            # Set the jump time reference.
            #
            # This function used to slide the time value further backward or
            # further forward to avoid the case where there's a Fact or Gap
            # longer than the jump time. In this case, as the user continues to
            # jump backward (or forward), they would keep seeing the same Fact.
            # - (lb): I had logic here that would sit in a while loop and adjust
            #   ref_time backward or forward by 1 day until the time reference
            #   was such that it would show a different Fact the next time the
            #   user jumped. But then I added the *count* feature, which allows
            #   users to jump by multiple days. And then this logic -- which did
            #   not feel well writ to begin with -- became untenable, given that
            #   the jump amount is no longer predictable.
            # E.g., consider the current Fact is the active Fact, now time is
            # 2020-04-12 09:30, and the user presses `5J` to go back five days,
            # ideally to 2020-04-07 09:30. Consider there's a Fact or Gap from
            # 2020-03-01 to 2020-04-07. Then the next `5J` will try
            # 2020-04-02 09:30, and the `5J` after that will try
            # 2020-03-28 09:30, etc. On each jump, the user sees the same Fact.
            # - (lb): So you can see how I wanted to make a better UX. However,
            #   in addition to the concerns mentioned above, there's another one.
            #   If we change the reference time, then operations are not
            #   transitive, e.g., user cannot `5J` from Fact A to B, and then
            #   run `5K` to return to Fact A.
            # - (lb): I think I've been able to solve this stumper by updating
            #   the bottom area Status message when the user jumps to show how
            #   many *days* the user jumped, and to also show the current
            #   reference time. That way, even if the current Fact does not
            #   change, the Status message will show something unique each
            #   time.
            self.jump_time_reference = ref_time

        # ***

        def debug_log_chosen_fact(
            ref_time, nearest_fact, fact_group, group_fact, store_fact,
        ):
            self.debug('ref_time: {}'.format(ref_time))
            self.debug(
                'near_group: {}'.format(fact_group.sorty_times)
            )
            self.debug(
                'nerst_fact: {}'.format(nearest_fact and nearest_fact.short)
            )
            self.debug(
                'group_fact: {}'.format(group_fact and group_fact.short)
            )
            self.debug(
                'store_fact: {}'.format(store_fact and store_fact.short)
            )
            if nearest_fact is None:
                chosen_from = 'neither'
            elif nearest_fact is group_fact:
                chosen_from = 'group'
            elif nearest_fact is store_fact:
                chosen_from = 'store'
            else:
                chosen_from = 'heh?'
            self.debug('- fact chosen from: {}'.format(chosen_from))

        # ***

        return _jump_to_fact_nearest()




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Facts Carousel"""

from gettext import gettext as _

__all__ = (
    'render_carousel_help',
    'NUM_HELP_PAGES',
)


def render_carousel_help():
    # (lb): Is this sacrilegious, importing from that which installed *us*?
    # - I'll at least move the import inside the method. Then if this library
    #   is used outside dob (h. unlikely), the blowup will be localised.
    from dob import get_version as get_version_dob

    # FIXME/2020-04-01: Revisit this. Some commands changed; some were never implemented!
    carousel_help = _(
        """ ┏━━━━━━━━━ NAVIGATION ━━━━━━━━┳━━━━ PROMPTS ━━━━┳━━━━━━━ NUDGE TIME ━━━━━━━┓
 ┃ → / ←   Next/Previous Fact  ┃ a  edit act@cat ┃ [  Sub 1 min. from end   ┃
 ┃ j / k     Same as → / ←     ┃ t  edit tags    ┃ ]  Add 1 minute to end   ┃
 ┃ ↑ / ↓   Move Cursor Up/Down ┣━━━━ $EDITOR ━━━━┫ ,  Sub 1 min. from start ┃
 ┃ h / l     Same as ↑ / ↓     ┃ d  edit descrip.┃ .  Add 1 minute to start ┃
 ┃ PgUp    Move Cursor Up/Down ┣━━━━ COPIERS ━━━━┫ {{  Sub 5 mins from end   ┃
 ┃  PgDn     by pageful        ┃  c-c   a@c+tags ┃ }}  Add 5 mins to end     ┃
 ┃ Home    Top of Description  ┃  c-v    paste   ┃ <  Sub 5 mins from start ┃
 ┃  End      Bottom of Desc.   ┃ A c-c  copy a@c ┃ >  Add 5 mins to start   ┃
 ┣━━━ JUMP JUMP JUMP AROUND ━━━┫ T c-c  cpy tags ┃ Ctrl-Shift →  Nudge both ┃
 ┃ J / K   Back 1 day / Fwd. 1 ┃ D c-c  cpy desc ┃ Ctrl-Shift ←  Nudge both ┃
 ┃ [n]J/K  Back/Fwd [n] days   ┣━━━━━━━━━ RELATIVE TIME ADJUSTMENTS ━━━━━━━━┫
 ┃ f       Jump to First Fact  ┃ +[n]m  Set end [n] mins. after start time  ┃
 ┃ F       Jump to Final Fact  ┃ +[n]h  Set end [n] hours after start time  ┃
 ┃ YYYYMMDD f  Jump to date    ┃ -[n]m  Set start [n] mins. before end time ┃
 ┣━━━━━━━┳━━━━━━━┳━━━━━━┳━━━━━━┫ -[n]h  Set start [n] hours before end time ┃
 ┃ c-s   ┃ c-q   ┃ c-z  ┃ c-y  ┃ [n][ [n]]  Sub/Add [n] minutes from/to end ┃
 ┃ Save  ┃ Exit  ┃ Undo ┃ Redo ┃ [n], [n].  Sub/Add [n] mins. from/to start ┃
 ┣━━━━━━━┻━━━━━━━┻━━━━━━┻━━━━━━┻━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┫
 ┃ c-e  Copy curr. Fact meta onto Active Fact ┃ o  Complete Fact and prompt ┃
 ┃ V    Use curr. Fact meta to start new Fact ┃      for new Fact act@cat   ┃
 ┣━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━┻━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┫
 ┃     Learn more tricks at       ┃  Copyright © 2018-2020 Landon Bouma     ┃
 ┃   https://dob.readthedocs.io   ┃  dob v.{dob_vers: <32} ┃
 ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┻━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛
        """
    ).format(
        dob_vers=get_version_dob()[:34],
    ).rstrip()
    return carousel_help


NUM_HELP_PAGES = 2




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Facts Carousel Header (Fact meta and diff)"""

import re

from prompt_toolkit.layout.containers import HSplit
from prompt_toolkit.widgets import Label

__all__ = (
    'ZoneStreamer',
)


class ZoneStreamer(object):
    """"""
    def __init__(self, carousel):
        self.carousel = carousel

    def standup(self):
        """"""
        def _standup():
            self.streamer_style = self.carousel.style_classes['streamer']
            self.zone_manager = self.carousel.zone_manager
            assemble_children()
            self.streamer_container = build_container()

        def assemble_children():
            self.children = []
            add_fact_banner()

        def add_fact_banner():
            add_interval_banner()

        # ***

        def add_interval_banner():
            self.interval_banner = Label(text='')
            self.children.append(self.interval_banner)

        # ***

        def build_container():
            streamer_container = HSplit(children=self.children)
            return streamer_container

        # ***

        _standup()

    # ***

    def rebuild_viewable(self):
        """"""
        self.refresh_all_children()
        return self.streamer_container

    # ***

    def refresh_all_children(self):
        self.refresh_interval()

    # ***

    def selectively_refresh(self):
        self.refresh_interval()

    # ***

    STREAMER_LINE_CLASS = 'class:streamer-line'

    RE_STYLE_HAS_CLASS = re.compile(r'\b{}\b'.format(STREAMER_LINE_CLASS))

    def refresh_interval(self):
        tod_humanize = self.zone_manager.facts_diff.edit_fact.time_of_day_humanize
        interval_text = tod_humanize(show_now=True)
        self.interval_banner.text = self.bannerize(interval_text)
        # (lb): Not sure why, but unlike process_style_rules, which sets
        # widget.formatted_text_control.style, here we set widget.window's
        # style instead.
        # (lb): Note that we set the style on refresh, and not standup, so
        # that classes are properly ordered. (The first element added is
        # 'class:label', which we want more specific classes to override.)
        # (lb): Reminder that match() starts at string beginning, so use search().
        match = ZoneStreamer.RE_STYLE_HAS_CLASS.search(self.interval_banner.window.style)
        if match is None:
            self.interval_banner.window.style += ' ' + ZoneStreamer.STREAMER_LINE_CLASS
        self.process_style_rules()

    def process_style_rules(self):
        # Register class:streamer[-line] styles.
        friendly_name = 'streamer'
        for suffix in ('', '-line'):
            # The label itself, 'streamer', includes the '╭─...' border,
            # so 3 rows of output are styled. If you specify the whole container,
            # 'streamer-line', the style will include the blank lines, one
            # each, above and below those 3 rows.
            self.carousel.process_style_rules(
                self.interval_banner,
                friendly_name + suffix,
            )

    # ***

    # Interval as currently formatted has max 53 chars, e.g.,
    MAX_INTERVAL_WIDTH = len('Fri 13 Jul 2018 ◐ 11:40 PM — 12:29 AM Sat 14 Jul 2018')

    def bannerize(self, text):
        def _bannerize():
            # There are a few ways to style the banner, including the ANSI border,
            # and optionally including the before and after newlines. By default,
            # the 'streamer' class from the current style is used to style the
            # banner and ANSI border; and the 'streamer-line' class is used to
            # style the same, and also the blank lines before and after. You can
            # also stylize the banner conditionally using the same two options,
            # 'streamer' and 'streamer-line', from a rules.conf rule set.
            bannerful = colorful_banner_town(text)
            parts = []
            # Inline style only applies to parts with text, so
            # using 'blank-line' style here would be fruitless.
            parts += [('', '\n')]
            parts += [(self.streamer_style, bannerful)]
            parts += [('', '\n')]
            return parts

        def colorful_banner_town(text):
            # MAYBE: (lb): We could go to the trouble of styling the ANSI border
            # separately than the text... but that seems a wee bit tedious to code.
            reps = self.carousel.avail_width - 1
            padded_text = '{0:<{1}}'.format(text, ZoneStreamer.MAX_INTERVAL_WIDTH)
            centered_text = '{0:^{1}}'.format(padded_text, reps)
            padded_hrule_top = '╭' + '─' * reps + '╮'
            padded_hrule_bot = '╰' + '─' * reps + '╯'
            banner = '{0}\n│{1}│\n{2}'.format(
                padded_hrule_top, centered_text, padded_hrule_bot,
            )
            return banner

        return _bannerize()




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

""""""

from prompt_toolkit.lexers import Lexer

__all__ = (
    'rainbow',
    'truncater',
    'wordwrapper',
    # Private:
    #  'BaseLexer',
)


class BaseLexer(Lexer):
    """"""
    def __init__(self):
        super(BaseLexer, self).__init__()
        self._content_width = None

    @property
    def content_width(self):
        return self._content_width

    @content_width.setter
    def content_width(self, content_width):
        self._content_width = content_width


def wordwrapper():
    class WordWrappingLexer(BaseLexer):
        """A very basic, primitive, "dumb", split-on-space line splitter."""
        def lex_document(self, document):
            splitdoc = []
            for line in document.lines:
                for chunk in line.split(' '):
                    splitdoc.append(chunk)

            def get_line(lineno):
                line = splitdoc[lineno]
                return [("#00ff55", line)]

            return get_line

    return WordWrappingLexer


def truncater():
    class TruncatingLexer(BaseLexer):
        @BaseLexer.content_width.setter
        def content_width(self, content_width):
            self._content_width = content_width
            self.dots_cnt = 3
            self.trunc_at = max(self.content_width - self.dots_cnt, 0)

        def lex_document(self, document):
            def get_line(lineno):
                line = document.lines[lineno]
                if self.content_width:
                    if len(line) > self.content_width:
                        line = line[:self.trunc_at]
                        line += '━' * self.dots_cnt
                return [("#00ff55", c) for c in line]

            return get_line

    return TruncatingLexer


def rainbow():
    from prompt_toolkit.styles.named_colors import NAMED_COLORS

    class RainbowLexer(BaseLexer):
        def lex_document(self, document):
            colors = list(sorted(NAMED_COLORS, key=NAMED_COLORS.get))

            def get_line(lineno):
                return [
                    (colors[i % len(colors)], c)
                    for i, c in enumerate(document.lines[lineno])
                ]

            return get_line

    return RainbowLexer




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Zone Details End Time Code (for easier diff-meld against Start Time Code)"""

from datetime import timedelta

from .exceptions import catch_action_exception

__all__ = (
    'ZoneDetails_TimeStart',
)


class ZoneDetails_TimeStart(object):
    """"""

    def __init__(self):
        super(ZoneDetails_TimeStart, self).__init__()

    def add_header_start_time(self):
        self.widgets_start = self.add_header_section(
            'start', 'start_fmt_local', editable=True,
            mouse_handler=self.header_time_start_mouse_handler,
        )

    def header_time_start_mouse_handler(self, mouse_event):
        self.zone_manager.toggle_focus_time_start(event=None)
        self.zone_manager.ppt_control_buffer_avoid_selection(
            self.widgets_start.text_area.control,
            mouse_event,
        )

    def refresh_time_start(self):
        self.refresh_val_widgets(self.widgets_start)

    # ***

    @catch_action_exception
    def edit_time_start(self, event=None, focus=True):
        if focus:
            edit_fact = self.zone_manager.facts_diff.edit_fact
            start_fmt_local = edit_fact.start_fmt_local
            self.widgets_start.orig_val = start_fmt_local
            # Styling/Rules note: PPT does not let you set format tuples
            # on TextArea text, i.e., text_area.text = [('class:foo', 'bar')]
            # would fail terribly. See elsewhere in the code where we are able
            # to use a SimpleLexer to stylize the input control.
            self.widgets_start.text_area.text = start_fmt_local
            self.edit_time_focus(self.widgets_start)
            return True
        else:
            return self.edit_time_leave(self.widgets_start)

    # ***

    def apply_edit_time_removed_start(self, edit_fact, passive=False):
        # Nothing ventured, nothing gained. Ignore deleted start. (We could
        # instead choose to do nothing, or choose to warn-tell user they're
        # an idiot and cannot clear the start time, or we could just behave
        # like a successful edit (by moving focus back to the matter (Fact
        # description) control) but not actually edit anything. Or we could
        # just do nothing. (User can tab-away and then we'll repopulate we
        # unedited time.)
        self.carousel.controller.affirm(edit_fact.start)
        self.widgets_start.text_area.text = edit_fact.start_fmt_local
        if passive:
            # User is tabbing away. We've reset the start, so let them.
            return True
        # User hit 'enter'. Annoy them with a warning.
        return False

    # ***

    def apply_edit_time_start(self, edit_fact, edit_time, verify_fact_times):
        if edit_fact.start == edit_time:
            # Nothing changed.
            return False
        edited_facts = [edit_fact]
        edits_manager = self.carousel.edits_manager
        # Make undoable.
        was_fact = edit_fact.copy()
        undoable_facts = [was_fact]
        # Prohibit completely shadowing other facts' time windows, but allow
        # changing one fact's times to shorten the times of prev or next fact.
        edit_prev = edits_manager.editable_fact_prev(edit_fact)
        best_time = edit_time
        if edit_prev and edit_prev.start and (best_time < edit_prev.start):
            # Adjust adjacent fact's time width to be no less that fact_min_delta.
            min_delta = int(self.carousel.controller.config['time.fact_min_delta'])
            # We do not want to make momentaneous facts, so use at least 1 minute.
            min_delta = max(1, min_delta)
            best_time = edit_prev.start + timedelta(minutes=min_delta)
            if edit_prev and edit_prev.end and (best_time > edit_prev.end):
                best_time = edit_prev.end
        was_time = edit_fact.start
        edit_fact.start = best_time
        # Verify edit_fact.start < edit_fact.end.
        if not verify_fact_times(edit_fact):
            edit_fact.start = was_time
            return False
        # If the edited time encroached on the neighbor, or if the neighbor
        # is an unedited gap fact, edit thy neighbor.
        if edit_prev:
            if (
                (edit_fact.start < edit_prev.end)
                or edit_prev.is_gap
            ):
                undoable_facts.append(edit_prev.copy())
                edit_prev.end = edit_fact.start
                edited_facts.append(edit_prev)
            if edit_fact.start == edit_prev.end:
                edit_fact.prev_fact = edit_prev
                edit_prev.next_fact = edit_fact
            else:
                edit_fact.prev_fact = None
                edit_prev.next_fact = None
        else:
            edit_fact.prev_fact = None
        edits_manager.add_undoable(undoable_facts, what='header-edit')
        edits_manager.apply_edits(*edited_facts)
        return True




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Key Binding Action Manager"""

from prompt_toolkit.application.application import _CombinedRegistry
from prompt_toolkit.key_binding.key_processor import KeyProcessor

from .interface_keys import (
    KeyBonder,
)
from .key_action_map import KeyActionMap

__all__ = (
    'ActionManager',
)


class ActionManager(object):
    """"""
    def __init__(self, carousel):
        self.carousel = carousel

    # ***

    def standup(self):
        self.key_action_map = KeyActionMap(self.carousel)
        self.key_bonder = KeyBonder(
            config=self.carousel.controller.config,
        )
        self.setup_key_bindings()

    def finalize_standup(self):
        self.key_bonder.print_warnings()

    # ***

    def _wire_keys(self, key_bindings):
        application = self.carousel.zone_manager.application
        previous_bindings = application.key_bindings
        application.key_bindings = key_bindings
        # A thing of beauty. What a hack job. (lb): I added this to support the
        # commando feature, but it may have always been missing (but I did not
        # notice because I wasn't throwing key combos at the time widgets); or
        # it might be that PTK 3.0 changes now demand it. Rebuild the key_processor,
        # otherwise the key_bindings just wired have no effect.
        application.key_processor = KeyProcessor(_CombinedRegistry(application))
        return previous_bindings

    def wire_keys_command_mode(self, key_bindings):
        # Set focus to something without an input control. Otherwise, if we leave
        # focus on time widget, the _CombinedRegistry will add keybindings for it!
        self.previous_control = self.carousel.zone_manager.layout.current_control
        self.carousel.zone_manager.layout.focus(
            self.carousel.zone_manager.zone_content.content
        )
        # SOOOOOO hacky. Love it! Disable cursor (otherwise there's a white
        # rectangle at 0,0 in the content window.)
        current_window = self.carousel.zone_manager.layout.current_window
        self.previous_hide_cursor = current_window.always_hide_cursor
        current_window.always_hide_cursor = lambda: True
        # Finally, wire the commando key bindings and handlers.
        self.previous_bindings = self._wire_keys(key_bindings)

    def unwire_keys_command_mode(self):
        # Enable content window to show cursor next time it's focused.
        current_window = self.carousel.zone_manager.layout.current_window
        current_window.always_hide_cursor = self.previous_hide_cursor
        self.previous_hide_cursor = None
        # Re-focus whatever widget was active before commando mode.
        self.carousel.zone_manager.layout.focus(self.previous_control)
        self.previous_control = None
        # Re-wire keys wired before commando mode.
        self._wire_keys(self.previous_bindings)
        self.previous_bindings = None

    # ***

    def wire_keys_normal(self):
        self._wire_keys(self.key_bindings_normal)

    def wire_keys_edit_time(self):
        self._wire_keys(self.key_bindings_edit_time)

    def wire_keys_modal(self):
        self._wire_keys(self.key_bindings_modal)

    def wire_keys_commando(self):
        self.wire_keys_command_mode(self.key_bindings_commando)

    def unwire_keys_commando(self):
        self.unwire_keys_command_mode()

    def wire_keys_delta_time(self):
        self.wire_keys_command_mode(self.key_bindings_delta_time)

    def unwire_keys_delta_time(self):
        self.unwire_keys_command_mode()

    # ***

    def setup_key_bindings(self):
        self.setup_key_bindings_shared()
        self.setup_key_bindings_normal()
        self.setup_key_bindings_edit_time()
        self.setup_key_bindings_modal()
        self.setup_key_bindings_commando()
        self.setup_key_bindings_delta_time()

    def setup_key_bindings_shared(self):
        bindings = []
        bindings += self.key_bonder.save_and_quit(self.key_action_map)
        bindings += self.key_bonder.widget_focus(self.key_action_map)

        self.key_bindings_shared = bindings

    def setup_key_bindings_normal(self):
        bindings = []
        bindings += self.key_bonder.normal(self.key_action_map)
        bindings += self.key_bonder.edit_fact(self.key_action_map)
        bindings += self.key_bonder.nudge_time(self.key_action_map)
        bindings += self.key_bonder.begin_delta_time(self.key_action_map)
        bindings += self.key_bonder.command_modifier(self.key_action_map)
        bindings += self.key_bonder.create_delete_fact(self.key_action_map)
        bindings += self.key_bonder.clipboard(self.key_action_map)
        bindings += self.key_bonder.undo_redo(self.key_action_map, 'content')
        bindings += self.key_bonder.shortcuts(self.key_action_map)
        bindings += self.key_bonder.begin_commando(self.key_action_map)
        bindings += self.key_bonder.custom_factoids(self.key_action_map)
        bindings += self.key_bindings_shared

        self.key_bindings_normal = self.key_bonder.make_bindings(bindings)

    def setup_key_bindings_edit_time(self):
        bindings = []
        bindings += self.key_bonder.edit_time(self.key_action_map)
        bindings += self.key_bonder.undo_redo(self.key_action_map, 'edit_time')
        bindings += self.key_bindings_shared

        self.key_bindings_edit_time = self.key_bonder.make_bindings(bindings)

    def setup_key_bindings_modal(self):
        bindings = []
        # None. Modal has its own for the basics.
        # SKIP: bindings += self.key_bindings_shared
        self.key_bindings_modal = self.key_bonder.make_bindings(bindings)

    def setup_key_bindings_commando(self):
        bindings = []
        bindings += self.key_bonder.going_commando(self.key_action_map)
        # SKIP: bindings += self.key_bindings_shared
        self.key_bindings_commando = self.key_bonder.make_bindings(bindings)

    def setup_key_bindings_delta_time(self):
        bindings = []
        bindings += self.key_bonder.going_delta_time(self.key_action_map)
        # SKIP: bindings += self.key_bindings_shared
        self.key_bindings_delta_time = self.key_bonder.make_bindings(bindings)




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Key Binding Wiring Manager"""

from gettext import gettext as _

from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys

from dob_bright.termio import dob_in_user_warning

from dob_prompt.prompters.interface_bonds import KeyBond

from ..config.custom_paste import DobViewerConfigCustomPaste
from ..config.json_sublist import json_load_sublisted

__all__ = (
    'KeyBonder',
)


# ***

class KeyBonder(object):

    def __init__(self, config):
        self.config = config
        self.errors = []
        self.keyed_factoids = {}

    # ***

    KEYBONDS_CFG_SECTION = 'editor-keys'

    def _key_bonds(
        self, action_map, action_name, config_name=None, config_section=None,
    ):
        """"""
        def __key_bonds():
            action, cfgval = resolve_action_cfgval()
            keybonds = build_bonds(action, cfgval)
            return keybonds

        def resolve_action_cfgval():
            editor_keys = config_section or KeyBonder.KEYBONDS_CFG_SECTION
            cfgname = config_name or action_name
            action = getattr(action_map, action_name)
            cfgval = self.config[editor_keys][cfgname]
            return action, cfgval

        def build_bonds(action, cfgval):
            # Note that the json loader discards any entry that would just
            # be the empty string (which the user can apply to a key binding to
            # effectively "disable" (render unreachable) the associated command).
            keycodes, errmsg = json_load_sublisted(action_name, cfgval)
            if errmsg is not None:
                return self.add_error_and_return_empty(errmsg)
            keybonds = [KeyBond(keycode, action=action) for keycode in keycodes]
            return keybonds

        return __key_bonds()

    def add_error_and_return_empty(self, errmsg):
        self.errors.append(errmsg)
        return []

    # ***

    def make_bindings(self, key_bonds):
        """"""
        def _make_bindings():
            key_bindings = KeyBindings()
            [add_binding(key_bindings, keyb) for keyb in key_bonds]
            return key_bindings

        def add_binding(key_bindings, keyb):
            try:
                add_binding_str_or_list(key_bindings, keyb)
            except Exception as err:
                self.errors.append(_(
                    'ERROR: Failed to add a key binding for ‘{}’: “{}”'
                    .format(keyb.action.__name__, str(err))
                ))

        def add_binding_str_or_list(key_bindings, keyb):
            # YOU: Toss an eager=True to the add if you think PTK is
            # hooking a binding. Or check your tmux config, mayhap.
            if isinstance(keyb.keycode, str):
                key_bindings.add(keyb.keycode)(keyb.action)
            else:
                key_bindings.add(*keyb.keycode)(keyb.action)

        return _make_bindings()

    # ***

    @property
    def date_separators(self):
        try:
            return self._date_separators
        except AttributeError:
            return self._load_date_separators()

    def _load_date_separators(self):
        editor_keys = KeyBonder.KEYBONDS_CFG_SECTION
        cfgname = 'date_separators'
        date_seps = self.config[editor_keys][cfgname]
        self._date_separators, errmsg = json_load_sublisted(cfgname, date_seps)
        if errmsg is not None:
            return self.add_error_and_return_empty(errmsg)
        return self._date_separators

    # ***

    def print_warnings(self):
        if not self.errors:
            return
        dob_in_user_warning('\n'.join(self.errors))
        self.errors = []

    # ***

    def widget_focus(self, action_map):
        key_bonds = []
        # Use the 'focus_next' config value as the key to wire
        # to the action_map.focus_next handler.
        key_bonds += self._key_bonds(action_map, 'focus_next')
        key_bonds += self._key_bonds(action_map, 'focus_previous')
        # Bindings to edit time are always available (and toggle focus when repeated).
        key_bonds += self._key_bonds(action_map, 'edit_time_start')
        key_bonds += self._key_bonds(action_map, 'edit_time_end')
        return key_bonds

    # ***

    def save_and_quit(self, action_map):
        key_bonds = []
        # Save Facts command is where you'd expect it.
        key_bonds += self._key_bonds(action_map, 'save_edited_and_live')
        key_bonds += self._key_bonds(action_map, 'save_edited_and_exit')
        # User can always real-quit, but prompted if edits.
        key_bonds += self._key_bonds(action_map, 'exit_command')
        # User can soft-cancel if they have not edited.
        key_bonds += self._key_bonds(action_map, 'exit_quietly')
        return key_bonds

    # ***

    def edit_time(self, action_map):
        key_bonds = []
        key_bonds += self._key_bonds(action_map, 'edit_time_enter')
        key_bonds += self._key_bonds(action_map, 'toggle_focus_description')
        # By default, PPT will add any key we don't capture to active widget's
        # buffer, but we'll override so we can ignore alpha characters.
        key_bonds += [KeyBond(Keys.Any, action=action_map.edit_time_any_key)]
        return key_bonds

    # ***

    def undo_redo(self, action_map, context):
        key_bonds = []
        key_bonds += self._key_bonds(
            action_map, 'undo_command_{}'.format(context), 'undo_command',
        )
        key_bonds += self._key_bonds(
            action_map, 'redo_command_{}'.format(context), 'redo_command',
        )
        return key_bonds

    # ***

    def normal(self, action_map):
        key_bonds = []

        key_bonds += self._key_bonds(action_map, 'rotate_help')
        key_bonds += self._key_bonds(action_map, 'dev_breakpoint')

        key_bonds += self._key_bonds(action_map, 'jump_fact_dec')
        key_bonds += self._key_bonds(action_map, 'jump_fact_inc')

        key_bonds += self._key_bonds(action_map, 'jump_day_dec')
        key_bonds += self._key_bonds(action_map, 'jump_day_inc')

        key_bonds += self._key_bonds(action_map, 'jump_rift_dec')
        key_bonds += self._key_bonds(action_map, 'jump_rift_inc')

        key_bonds += self._key_bonds(action_map, 'jump_fact_first')
        key_bonds += self._key_bonds(action_map, 'jump_fact_final')

        key_bonds += self._key_bonds(action_map, 'cursor_up_one')
        key_bonds += self._key_bonds(action_map, 'cursor_down_one')

        key_bonds += self._key_bonds(action_map, 'scroll_up')
        key_bonds += self._key_bonds(action_map, 'scroll_down')
        key_bonds += self._key_bonds(action_map, 'scroll_top')
        key_bonds += self._key_bonds(action_map, 'scroll_bottom')

        # FIXME/BACKLOG: Search feature. E.g., like Vim's /:
        #   KeyBond('/', action=zone_lowdown.start_search),
        # FIXME/BACKLOG: Filter feature.
        #   (By tag; matching text; dates; days of the week; etc.)

        # (lb): Not every Vim key needs to be mapped, e.g.,
        #  KeyBond('M', action=action_map.jump_fact_midpoint),
        # seems capricious, i.e., why implement if not just because we can?

        return key_bonds

    # ***

    def edit_fact(self, action_map):
        key_bonds = []
        key_bonds += self._key_bonds(action_map, 'edit_fact')
        key_bonds += self._key_bonds(action_map, 'edit_actegory')
        key_bonds += self._key_bonds(action_map, 'edit_description')
        key_bonds += self._key_bonds(action_map, 'edit_tags')
        return key_bonds

    # ***

    def nudge_time(self, action_map):
        key_bonds = []
        key_bonds += self._key_bonds(action_map, 'edit_time_decrement_start')
        key_bonds += self._key_bonds(action_map, 'edit_time_increment_start')
        key_bonds += self._key_bonds(action_map, 'edit_time_decrement_end')
        key_bonds += self._key_bonds(action_map, 'edit_time_increment_end')
        key_bonds += self._key_bonds(action_map, 'edit_time_decrement_both')
        key_bonds += self._key_bonds(action_map, 'edit_time_increment_both')
        key_bonds += self._key_bonds(action_map, 'edit_time_decrement_start_5min')
        key_bonds += self._key_bonds(action_map, 'edit_time_increment_start_5min')
        key_bonds += self._key_bonds(action_map, 'edit_time_decrement_end_5min')
        key_bonds += self._key_bonds(action_map, 'edit_time_increment_end_5min')
        return key_bonds

    # ***

    def command_modifier(self, action_map):
        key_bonds = []
        key_bonds += self._key_bonds(action_map, 'allow_time_gap')
        key_bonds += [KeyBond(Keys.Any, action=action_map.command_modifier_any_key)]
        key_bonds += [KeyBond('c-h', action=action_map.backspace_command_modifier)]
        return key_bonds

    # ***

    def create_delete_fact(self, action_map):
        return []  # FIXME/2020-04-11: Implement or get off the pot!

        # FIXME/2020-04-11: Not implemented:
        key_bonds = []
        key_bonds += self._key_bonds(action_map, 'fact_split')
        key_bonds += self._key_bonds(action_map, 'fact_erase')
        key_bonds += self._key_bonds(action_map, 'fact_merge_prev')
        key_bonds += self._key_bonds(action_map, 'fact_merge_next')
        return key_bonds

    # ***

    def clipboard(self, action_map):
        key_bonds = []
        key_bonds += self._key_bonds(action_map, 'fact_copy_fact')
        key_bonds += self._key_bonds(action_map, 'fact_cut')
        key_bonds += self._key_bonds(action_map, 'fact_paste')
        key_bonds += self._key_bonds(action_map, 'fact_copy_activity')
        key_bonds += self._key_bonds(action_map, 'fact_copy_tags')
        key_bonds += self._key_bonds(action_map, 'fact_copy_description')
        return key_bonds

    # ***

    def shortcuts(self, action_map):
        key_bonds = []
        key_bonds += self._key_bonds(action_map, 'copy_complete_and_paste_active')
        key_bonds += self._key_bonds(action_map, 'copy_complete_and_paste_new')
        key_bonds += self._key_bonds(action_map, 'complete_and_prompt_new')
        return key_bonds

    # ***

    FACTOID_CFG_SECTION = 'custom-paste'
    FACTOID_CFG_MAPPING = 'mapping_'
    FACTOID_CFG_FACTOID = 'factoid_'

    def custom_factoids(self, action_map):
        self.keyed_factoids = {}
        key_bonds = []
        custom_paste = KeyBonder.FACTOID_CFG_SECTION
        for postfix in range(1, DobViewerConfigCustomPaste._innercls.A_PERFECT_NUMBER):
            mapping_name = '{}{}'.format(KeyBonder.FACTOID_CFG_MAPPING, postfix)
            custom_bonds = self._key_bonds(
                action_map,
                'custom_factoid_paste',
                config_name=mapping_name,
                config_section=custom_paste,
            )
            if not custom_bonds:
                # No custom mapping at this postfix.
                continue
            custom_bond = custom_bonds[0]
            factoid_name = '{}{}'.format(KeyBonder.FACTOID_CFG_FACTOID, postfix)
            custom_factoid = self.config[custom_paste][factoid_name]
            if not custom_factoid:
                self.errors.append(_(
                    'ERROR: Custom key mapping ‘{}’ has no matching factoid: “{}”'
                    .format(mapping_name, factoid_name)
                ))
            else:
                self.keyed_factoids[custom_bond.keycode] = custom_factoid
                key_bonds += [custom_bond]
        return key_bonds

    # ***

    def begin_commando(self, action_map):
        # The Colon Commando! Because (by default) type ':' then command + 'ENTER'.
        # I.e., a Vim-like command mode.
        key_bonds = []
        key_bonds += self._key_bonds(action_map, 'begin_commando')
        return key_bonds

    def going_commando(self, action_map):
        key_bonds = []
        # So Ctrl-c is not user configurable. Go configure.
        key_bonds += [KeyBond('c-c', action=action_map.cancel_commando)]
        key_bonds += [KeyBond(Keys.Any, action=action_map.parts_commando)]
        key_bonds += [KeyBond('c-h', action=action_map.backspace_commando)]
        key_bonds += self._key_bonds(action_map, 'final_commando')
        return key_bonds

    # ***

    def begin_delta_time(self, action_map):
        key_bonds = []
        key_bonds += self._key_bonds(action_map, 'begin_delta_time_start')
        key_bonds += self._key_bonds(action_map, 'begin_delta_time_end')
        return key_bonds

    def going_delta_time(self, action_map):
        key_bonds = []
        # Any non-recognizable key cancels the binding, or Ctrl-C, which is
        # otherwise not handled unless we explicitly do so.
        key_bonds += [KeyBond('c-c', action=action_map.cancel_delta_time)]
        key_bonds += [KeyBond(Keys.Any, action=action_map.parts_delta_time)]
        key_bonds += [KeyBond('c-h', action=action_map.backspace_delta_time)]
        key_bonds += self._key_bonds(action_map, 'allow_time_gap')
        key_bonds += self._key_bonds(action_map, 'final_delta_time_apply')
        key_bonds += self._key_bonds(action_map, 'final_delta_time_minutes')
        key_bonds += self._key_bonds(action_map, 'final_delta_time_hours')
        return key_bonds

